#ifndef __CDCRAWDATA__
#define __CDCRAWDATA__

//*************************************************************************
//* ----------------
//* CDCRawData Class
//* ----------------
//* (Description)
//*	This set of classes reads CDC raw data for Lorentz Angle 
//*	measurments and unpacks them into root objects.
//*	When a root file is opened in the read/write mode, the root
//*	objects are stored in it.  If the root file is read-only,
//*	the data are obtained from it.
//*	The package contains the following classes:
//*	The following parameters can be specified in "jsf.conf". 
//*
//*	    CDCRawData.RFNFormat:       printf format to construct file name
//*	    CDCRawData.DoPlots:         (TRUE/FASLE)
//* (Requires)
//*	TObject:
//*	TClonesArray:
//*	CDCUniversalRecordHeader:
//*	CDCRawADCDatum:
//* (Provides)
//*	CDCHardwareMap		: hardware map
//*	CDCBeginRunBuf		: begin run buffer
//*	CDCRawDataBuf		: raw data buffer
//*	CDCEndRunBuf		: end run buffer
//*	CDCRawData		: steering class
//* (Update Record)
//*	1999/06/02  K.Fujii	Original version.
//*	2000/05/17  K.Fujii	Implementation for new format (not tested)
//*	2001/05/27  K.Fujii	Modification for new branch scheme
//*				started from ROOT v.3.01.
//*	2001/09/14  K.Fujii	Use automatic schema evolution.
//*************************************************************************

#include "TObject.h"
#include "TArrayI.h"
#include "TObjString.h"
#include "TClonesArray.h"
#include "TDatime.h"

#include "CDCTypeDef.h"
#include "CDCUniversalRecordHeader.h"
#include "CDCRawGPIBDatum.h"
#include "CDCRawScalerDatum.h"
#include "CDCRawADCDatum.h"
#include "DynArray3.h"

class CDCHardwareMap : public DynArray3 {
public:
  CDCHardwareMap() {}
  virtual ~CDCHardwareMap() {}

  ClassDef(CDCHardwareMap, 1) // CDC HardwareMap
};

class CDCChannel : public TObject {
friend class CDCBeginRunBuf;
public:
  CDCChannel(Int_t id = 0, Int_t ch = 0) : fID(id), fDetChNo(ch) {}
  virtual ~CDCChannel() {}
private:
  inline Int_t GetID   () const { return fID;      }
  inline Int_t GetDetCh() const { return fDetChNo; } 

private:
  Int_t fID;		// detector ID
  Int_t fDetChNo;	// detector channel No

  ClassDef(CDCChannel, 1) // FADC channel map
};

// ----------------
// Begin Run Buffer
// ----------------

class CDCRawData;
class CDCBeginRunBuf : public TObject {
public:
  CDCBeginRunBuf();
  virtual ~CDCBeginRunBuf();

  Bool_t ReadFile(FILE *fd);
  void   Unpack(BRR_t data[]);
  
  inline CDCUniversalRecordHeader *GetHeader  ()       { return &fHeader;   }
  inline Int_t			   GetDateTime() const { return  fDateTime; }
  inline TObjString 	     	  *GetComment ()       { return &fComment;  }
  void				   DebugPrint ()       {}
  void   EChannelToDChannel(Int_t cr, Int_t st, Int_t ch, 
  			    Int_t &id, Int_t &dch);

private:
  void   UnpackDateTime(BRR_t data[]);
  void   UnpackHardwareMap(BRR_t data[]);
  void   UnpackThresholdData(BRR_t data[]);


private:
  CDCUniversalRecordHeader  fHeader;	// Header
  Int_t		   	    fDateTime ;	// Date and time
  CDCHardwareMap	    fMap;	// (sta#,ch#) --> detector map
  TArrayI          	    fThresh;	// threshold data 
  TObjString		    fComment;	// Comment string

  ClassDef(CDCBeginRunBuf, 1) // CDC BeginRun record
};

// ---------------
// Raw Data Buffer
// ---------------

class CDCRawDataBuf : public JSFEventBuf {
private:
  CDCUniversalRecordHeader  fHeader; 		// Universal Record Header
  Len_t		    	    fNGPIBData;		// No. GPIB data units
  Len_t		    	    fNScalerData;	// No. Scaler data units
  Len_t		    	    fNADCData;		// No. ADC data units
  TClonesArray		   *fGPIBData;		// Array of GPIB Data
  TClonesArray		   *fScalerData;	// Array of Scaler Data
  TClonesArray		   *fADCData;		// Array of ADC Data
  Int_t 		    fNonFADCIndex; 	// index of 1st nonFADC datum
public:
  CDCRawDataBuf(const char *name="CDCRawDataBuf",
		const char *title="CDC raw data", CDCRawData *mod=0);
  CDCRawDataBuf(CDCRawData *mod, const char *name="CDCRawDataBuf",
		const char *title="CDC raw data");
  virtual ~CDCRawDataBuf();

  Bool_t ReadFile(FILE *fd);
  void   Unpack(ADC_t data[]);
  inline CDCUniversalRecordHeader * GetHeader()   { return &fHeader;      }
  inline Len_t		  GetNGPIBData   () const { return  fNGPIBData;   }
  inline Len_t		  GetNScalerData () const { return  fNScalerData; }
  inline Len_t		  GetNADCData    () const { return  fNADCData;    }
  inline TClonesArray	* GetGPIBData    () const { return  fGPIBData;    }
  inline TClonesArray	* GetScalerData  () const { return  fScalerData;  }
  inline TClonesArray   * GetADCData     () const { return  fADCData;     }
  inline Int_t 		  GetNonFADCIndex() const { return  fNonFADCIndex;}

  ClassDef(CDCRawDataBuf, 2) // CDC raw data buffer
};

// --------------
// End Run Buffer
// --------------

class CDCEndRunBuf : public TObject {
private:
  CDCUniversalRecordHeader  fHeader;	// Header 
  Int_t		   	    fDateTime ;	// Date and time
  TObjString		    fComment;	// Comment string
public:
  CDCEndRunBuf() {}
  virtual ~CDCEndRunBuf() {}

  Bool_t ReadFile(FILE *fd);
  void   Unpack(ERR_t data[]);
  
  inline CDCUniversalRecordHeader * GetHeader  () { return &fHeader;   }
  inline Int_t 	     	     	    GetDateTime() const { return  fDateTime; }
  inline TObjString 	     	  * GetComment ()       { return &fComment;  }
  void 			     	    DebugPrint () {}

  ClassDef(CDCEndRunBuf, 1) // CDC EndRun record
};

// -----------------
// Raw Data Analysis
// -----------------

class CDCRawData : public JSFModule {
private:
   CDCBeginRunBuf    *fBeginBuf;	//  Begin Run Record
   Char_t   	     *fRFNFormat;	//! Input file name format
   FILE     	     *fd;		//! File descriptor for the raw data

public:
   CDCRawData(const Char_t *name="CDCRawData",
	      const Char_t *title="Get CDC Raw data");
   virtual ~CDCRawData();

   void MakeBranch(TTree *tree);
   void SetBranch(TTree *tree);

   Bool_t BeginRun(Int_t nrun);
   Bool_t Process (Int_t ev);
   void   DebugPrint() {}

   inline CDCBeginRunBuf    * GetBeginRunBuf() const { return fBeginBuf; }

   ClassDef(CDCRawData, 1) // Read CDC raw data analysis
};

#endif


