//*************************************************************************
//* ----------------
//* CDCRawData Class
//* ----------------
//* (Description)
//*	This set of classes reads CDC raw data for Lorentz Angle 
//*	measurments and unpacks them into root objects.
//*	When a root file is opened in the read/write mode, the root
//*	objects are stored in it.  If the root file is read-only,
//*	the data are obtained from it.
//*	The package contains the following classes:
//*	The following parameters can be specified in "jsf.conf". 
//*
//*	    CDCRawData.RFNFormat:       printf format to construct file name
//*	    CDCRawData.DoPlots:         (TRUE/FASLE)
//* (Requires)
//*	TObject:
//*	TClonesArray:
//*	CDCUniversalRecordHeader:
//*	CDCRawGPIBDatum:
//*	CDCRawScalerDatum:
//*	CDCRawADCDatum:
//* (Provides)
//*	CDCHardwareMap		: hardware map
//*	CDCRawDataBuf		: raw data buffer
//*	CDCBeginRunBuf		: begin run buffer
//*	CDCEndRunBuf		: end run buffer
//*	CDCRawData			: steering class
//* (Update Record)
//*	1999/06/02  K.Fujii	Original version.
//*	2000/05/17  K.Fujii	Implementation for new format (not tested)
//*     2001/05/27  K.Fujii     Modification for new branch scheme
//*                             started from ROOT v.3.01.
//*	2001/09/14  K.Fujii     Use automatic schema evolution.
//*				Reverted branch name to CDCRawData-EventBuf
//*				to use legacy unpacked data.
//*	2001/11/15  K.Hoshina   Added CDCRawGPIBDatum and CDCRawScalerDatum.
//*	2002/02/10  K.Fujii	Protect against invalid access to 
//* 				non-existent entires in Hardware Map.
//*************************************************************************

#include "JSFSteer.h"
#include "CDCRawData.h"
#include "TRandom.h"
#include <iostream.h>

ClassImp(CDCChannel)
ClassImp(CDCHardwareMap)
ClassImp(CDCBeginRunBuf)
ClassImp(CDCRawDataBuf)
ClassImp(CDCEndRunBuf)
ClassImp(CDCRawData)

#define __DEBUG__

//___________________________________________________________________________
// ----------------
// Begin Run Buffer
// ----------------
//
CDCBeginRunBuf::CDCBeginRunBuf() : fDateTime(0) {}

//___________________________________________________________________________
CDCBeginRunBuf::~CDCBeginRunBuf() {}

//___________________________________________________________________________
Bool_t CDCBeginRunBuf::ReadFile(FILE *fd) {
    fHeader.ReadFile(fd);
    if ( fHeader.GetRecordType() != kRecTypeBRR ) return kFALSE;
    Len_t   nw   = fHeader.GetDataLength();
    cerr << "nw = " << nw << endl;
    BRR_t  *data = new BRR_t[nw];
    Len_t   ne   = fread(data, sizeof(BRR_t), nw, fd);
    if ( ne != nw ) {
        Error("BeginRunBuf","Failed to read BRR: nw =%d",ne);
        delete [] data;
        return kFALSE;
    }
    fHeader.DebugPrint();
    Unpack(data);
    delete [] data;
    return kTRUE;
}

//___________________________________________________________________________
void CDCBeginRunBuf::Unpack(BRR_t data[])
{
    Int_t fmtid = fHeader.GetFormatID();
    Len_t nw    = fHeader.GetDataLength();
    Int_t ic;
    switch (fmtid) {
        case 0: {				// Old 98 format
            ic = 3;
            break;
        }
        case 1: 				// New format
        case 2: {				// New format
        
            Int_t id = 0;
            UnpackDateTime(&data[id]);
            Int_t im = 2;
            UnpackHardwareMap(&data[im]);
            Int_t it = im + data[im];
            UnpackThresholdData(&data[it]);
            ic = it + data[it];
            break;
        }
        default: { 
            cerr << "Format ID = " << fmtid
                 << " is not supported" << endl;
            return;
        }
    }
    Char_t *comment =  new Char_t[2*(nw-ic)];
    sscanf((Char_t *)&data[ic],"%s",comment);
    fComment.SetString(comment);
    cout << "--------------------" << endl
         << " Comment of the Run" << endl
         << "--------------------" << endl;
    cout << fComment.GetString() << endl;
}

//___________________________________________________________________________
void CDCBeginRunBuf::UnpackDateTime(BRR_t data[])
{
    UShort_t dum[2];
    dum[0] = data[0];
    dum[1] = data[1];
#ifndef __LITTLE_ENDIAN__
    dum[0] = (dum[0] << 8) | (dum[0] >> 8);
    dum[1] = (dum[1] << 8) | (dum[1] >> 8);
    UInt_t dum4 = dum[1];
    fDateTime = (dum4 << 16) | dum[0];
#else
    UInt_t dum4 = dum[0];
    fDateTime = (dum4 << 16) | dum[1]; 
#endif
}

//___________________________________________________________________________
void CDCBeginRunBuf::UnpackHardwareMap(BRR_t data[])
{
    Len_t nw;
    Int_t i;
#ifndef __LITTLE_ENDIAN__
    nw = (UShort_t)(data[0] << 8) | (data[0] >> 8);
    for (i=0; i<nw; i++) {
    	data[i] = (data[i] << 8) | (data[i] >> 8);
    }
#else
    nw = data[0];
#endif
    UShort_t dum1, dum2;
    UShort_t nch = nw/2;
    for (i=0; i<nch; i++) {
        dum1 = data[2*i+1];
        dum2 = data[2*i+2];
        Int_t crtno = (UShort_t)(dum1 >> 9);
        Int_t stano = (UShort_t)((UShort_t)(dum1 << 7) >> 11);
        Int_t chano = (UShort_t)((UShort_t)(dum1 << 12) >> 12);
        Int_t id    = (UShort_t)(dum2 >> 14);
        Int_t detch = (UShort_t)((UShort_t)(dum2 << 2) >> 2) - 1;
        CDCChannel *chap = new CDCChannel(id,detch);
        fMap.Set(crtno,stano,chano) = chap;

#if 1	
	cerr << "Hardware Map ID, st#, ch#, dtc# = " 
	     << id << " " << stano << " " << chano << " " << detch << endl;
#endif        
        
    }
}

//___________________________________________________________________________
void CDCBeginRunBuf::UnpackThresholdData(BRR_t data[])
{
    Int_t i;
#ifndef __LITTLE_ENDIAN__
    Int_t nw = (UShort_t)(data[0] << 8) | (data[0] >> 8);
    for (i=0; i<nw; i++) {
    	data[i] = (data[i] << 8) | (data[i] >> 8);
    }
#else
    Int_t nw = data[0];
#endif
    fThresh.Set(nw-1);
    for (i=0; i<nw-1; i++) {
        fThresh[i] = (UShort_t)data[i+1];
    }
}
//___________________________________________________________________________
void CDCBeginRunBuf::EChannelToDChannel(Int_t cr, Int_t st, Int_t ch, 
  			    		Int_t &id, Int_t &dch)
{
    CDCChannel *chap = (CDCChannel *)fMap(cr,st,ch);
    if (chap) {
        id  = chap->GetID();
        dch = chap->GetDetCh();
    } else {
        cerr << ">>>>>> Error in CDCBeginRunBuf::EChannelToDChannel..." << endl
             << " (cr,st,ch) = (" << cr << "," << st << "," << ch << ")"
             << " is non-existent!" << endl
             << " ----> id and dch set to 9999." << endl;
        id  = 9999;
        dch = 9999;
    }
}


//___________________________________________________________________________
// ---------------
// Raw Data Buffer
// ---------------
//
CDCRawDataBuf::CDCRawDataBuf(const char *name, const char *title, 
	CDCRawData *module) : JSFEventBuf(name, title, (JSFModule*)module),
			      fNGPIBData(0), fNScalerData(0), fNADCData(0), 
			      fGPIBData(0), fScalerData(0), fADCData(0),
			      fNonFADCIndex(-999999) {}

//___________________________________________________________________________
CDCRawDataBuf::CDCRawDataBuf(CDCRawData *module, const char *name, 
	const char *title) : JSFEventBuf(name, title, (JSFModule*)module),
			     fNGPIBData(0), fNScalerData(0), fNADCData(0), 
			     fGPIBData(0), fScalerData(0), fADCData(0),
			     fNonFADCIndex(-999999) {}

//___________________________________________________________________________
CDCRawDataBuf::~CDCRawDataBuf()
{
    if ( fGPIBData )   delete fGPIBData;
    if ( fScalerData ) delete fScalerData;
    if ( fADCData )    delete fADCData;
}

//___________________________________________________________________________
Bool_t CDCRawDataBuf::ReadFile(FILE *fd)
{
  CDCUniversalRecordHeader header;
  for (Int_t i=0; i<3; i++) {
    if ( !fHeader.ReadFile(fd) ) return kFALSE;
    switch ( fHeader.GetRecordType() ) {
	case kRecTypeBRR:
	case kRecTypeERR:
		return kFALSE;
    }
    Int_t recID = fHeader.GetRecordType(); 
    Len_t   nw  = fHeader.GetDataLength();
    ADC_t  *data = new ADC_t[nw];
    Len_t   ne   = fread(data, sizeof(ADC_t), nw, fd);
    if ( ne != nw ) {
        Error("RawDataBuf","Failed to read Event Data: nw =%d",ne);
        delete [] data;
        return kFALSE;
    }
    SetEventNumber(fHeader.GetEventNumber());
    switch ( recID ) {
	case kRecTypeADC:
		header = fHeader;
   		CDCRawADCDatum::SetFormat(fHeader);
    		fNADCData = nw/CDCRawADCDatum::GetRawLength();
    		Unpack(data);
		break;
	case kRecTypeGPIB:
    		CDCRawGPIBDatum::SetFormat(fHeader);
    		fNGPIBData = nw/CDCRawGPIBDatum::GetRawLength();
    		Unpack(data);
		break;
	case kRecTypeScaler:
    		CDCRawScalerDatum::SetFormat(fHeader);
    		fNScalerData = nw/CDCRawScalerDatum::GetRawLength();
    		Unpack(data);
		break;
	default:
		if (recID > 0) {
        		Error("RawDataBuf","Record Type not supported: Type=%d",recID);
		}
    		delete [] data;
    		return kFALSE;
    }
    delete [] data;
    //  >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>   Oo. >')))<
    if (fHeader.GetFormatID() != 2) break;
    //  >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>   Oo. >')))<
  }
  fHeader = header;
  return kTRUE;
}

//___________________________________________________________________________
void CDCRawDataBuf::Unpack(ADC_t data[])
{
    switch ( fHeader.GetRecordType() ) {
	case kRecTypeGPIB: {
    		if ( !fGPIBData ) fGPIBData = new TClonesArray("CDCRawGPIBDatum",fNGPIBData);
    		TClonesArray & gpibdata = *fGPIBData;
    		Len_t lgpib = CDCRawGPIBDatum::GetRawLength();
    		for ( Int_t i = 0; i < fNGPIBData; i++ ) {
        		new(gpibdata[i]) CDCRawGPIBDatum(&data[lgpib*i]);
    		}
		break;
	}
	case kRecTypeScaler: {
    		if ( !fScalerData ) fScalerData = new TClonesArray("CDCRawScalerDatum",fNScalerData);
    		TClonesArray & scalerdata = *fScalerData;
    		Len_t lscaler = CDCRawScalerDatum::GetRawLength();
    		for ( Int_t i = 0; i < fNScalerData; i++ ) {
        		new(scalerdata[i]) CDCRawScalerDatum(&data[lscaler*i]);
    		}
		break;
	}
	case kRecTypeADC: {
		CDCBeginRunBuf *br = ((CDCRawData *)fModule)->GetBeginRunBuf();
    		if ( !fADCData ) fADCData = new TClonesArray("CDCRawADCDatum",fNADCData);
    		TClonesArray & adcdata = *fADCData;
    		Len_t ladc = CDCRawADCDatum::GetRawLength();
		fNonFADCIndex = -999999;
    		for ( Int_t i = 0; i < fNADCData; i++ ) {
        		new(adcdata[i]) CDCRawADCDatum(&data[ladc*i]);
			if (fNonFADCIndex < 0) {
				Int_t cr = 0; // there is only 1 crate now
				Int_t st = ((CDCRawADCDatum *)adcdata[i])->GetStationNo();
				Int_t ch = ((CDCRawADCDatum *)adcdata[i])->GetChannelNo();
				Int_t id = 0, a;
				br->EChannelToDChannel(cr,st,ch,id,a);
				if (id) fNonFADCIndex = i;
			}
    		}
	}
    }
}
//___________________________________________________________________________
void CDCRawDataBuf::Streamer(TBuffer &R__b)
{
   // Stream an object of class CDCRawDataBuf.

   if (R__b.IsReading()) {
      UInt_t R__s, R__c;
      Version_t R__v = R__b.ReadVersion(&R__s, &R__c);
      if (R__v > 1) {
        CDCRawDataBuf::Class()->ReadBuffer(R__b, this, R__v, R__s, R__c);
        return;
      }
   } else {
      CDCRawDataBuf::Class()->WriteBuffer(R__b, this);
   }
}

//___________________________________________________________________________
// --------------
// End Run Buffer
// --------------
//
Bool_t CDCEndRunBuf::ReadFile(FILE *fd) {
    fHeader.ReadFile(fd);
    if ( fHeader.GetRecordType() != kRecTypeERR ) return kFALSE;
    Len_t  nw   = fHeader.GetDataLength();
    ERR_t *data = new ERR_t[nw];
    Len_t  ne = fread(data, sizeof(ERR_t), nw, fd);
    if ( ne != nw ) {
        Error("BeginRunBuf","Failed to read ERR: nw =%d",ne);
        delete [] data;
        return kFALSE;
    }
    Unpack(data);
    delete [] data;
    return kTRUE;
}
//___________________________________________________________________________
void CDCEndRunBuf::Unpack(ERR_t data[])
{
}

//___________________________________________________________________________
// -----------------
// Raw Data Analysis
// -----------------
//
CDCRawData::CDCRawData(const char *name, const char *title)
       : JSFModule(name,title)
{
    cerr << " constructor for CDCRawData is called..." << endl;
    fBeginBuf = new CDCBeginRunBuf();
    fEventBuf = new CDCRawDataBuf(this);

    Len_t len = strlen( gJSF->Env()->GetValue("CDCRawData.RFNFormat",
                         "/proj/cdc/data/baby/2000/cosm/Run%5.5d.dat") );
    fRFNFormat = new Char_t[len];
    strcpy( fRFNFormat,  gJSF->Env()->GetValue("CDCRawData.RFNFormat",
                         "/proj/cdc/data/baby/2000/cosm/Run%5.5d.dat") );
    fd = NULL;
    SetBufferSize(2000000);  // buffer size for event data.
}

//___________________________________________________________________________
CDCRawData::~CDCRawData()
{
    cerr << " CDCRawData will be deleted...fEventBuf is " 
    	 << (Int_t)fEventBuf << endl;
    delete fBeginBuf;
    delete fEventBuf;
    delete fRFNFormat;
    if(fd) fclose(fd);
}

//___________________________________________________________________________
Bool_t CDCRawData::BeginRun(Int_t nrun)
{
    cout << " BeginRun is called for run " << nrun << endl;
        
    // Open Raw Data Root File
   
    if ( ! fFile->IsWritable() ) {
    	TDirectory *last = gDirectory;
    	Char_t dir[100];
    	sprintf(dir,"/conf/begin%5.5i",nrun);
    	fFile->cd(dir);
    	cout << " current file name is " << fFile->GetPath() << endl;
    	gDirectory->pwd();
    	fBeginBuf->Read("CDCBeginRunBuf");
	cout << " 	BRR read from it " << endl;
    	last->cd();
        return kTRUE;
    }

    // Open Raw Data File
    
    Len_t ln = strlen(fRFNFormat);
    Char_t *infile = new Char_t[ln+10];
    sprintf(infile,fRFNFormat,nrun);
    cout << " Going to open a file " << infile << endl;

    fd = fopen(infile,"rb");

    if ( fd == NULL ) {
        Error("BeginRun","Failed to open %s.",infile);
        delete [] infile;
        return kFALSE;
    }
    delete [] infile;
    
    // Read and Unpack Begin Run Record
    
    cout << " Going to unpack BRR " << endl;
    if ( ! fBeginBuf->ReadFile(fd) ) return kFALSE;
    
    Int_t irun = fBeginBuf->GetHeader()->GetRunNumber();
    cout << " BRR unpacked " << endl;
  
    // output standard information
    
    SetModuleStatus(kBeginRun);
    SetRunNumber(irun);
    ConfDirectory();
    
    // Save begin run record
    
    TDirectory *last = gDirectory;
    Char_t dir[100];
    sprintf(dir,"/conf/begin%5.5i",nrun);
    fFile->cd(dir);
    cout << " current file name is " << fFile->GetPath() << endl;

    gDirectory->pwd();
    fBeginBuf->Write("CDCBeginRunBuf");
    last->cd();

    return kTRUE;
}

//___________________________________________________________________________
Bool_t CDCRawData::Process(Int_t ev)
{
    CDCRawDataBuf *buf = (CDCRawDataBuf*)fEventBuf;
    if ( fFile->IsWritable() ) {
REPEAT:
        while (!buf->ReadFile(fd)) {
            Int_t id = buf->GetHeader()->GetRecordType();
            switch (id) {
                case kRecTypeERR: {
                    cout << "ERR encountered " << endl;
                    SetModuleStatus(kEndRun);
                    return kFALSE;
                }
                case -9999: {
                    cout << "EOF encountered " << endl;
                    return kFALSE;
                }
                default: {
                    cout << " Obtained record type is " << id << endl;
                }
            }
        }
    
        if (gJSF->GetRunNumber() != buf->GetHeader()->GetRunNumber()) {
            cerr << "????? ERROR ????? Run number inconsistency" << endl;
            buf->GetHeader()->DebugPrint();
            goto REPEAT;
        }
    }
    buf->GetHeader()->DebugPrint();

    return kTRUE;
}

//___________________________________________________________________________
void CDCRawData::MakeBranch(TTree *tree)
{   
   fTree=tree;
   if( fEventBuf && fMakeBranch ) {
#if ROOT_VERSION_CODE >= ROOT_VERSION(3,1,3)
     Int_t split=-1;
#else
     Int_t split=0;
#endif
     Int_t bsize= 200000;
//
// 2001/09/14  Reverted to old branch name to use legacy unpacked data.
#if 0
     tree->Branch(fEventBuf->GetName(), fEventBuf->ClassName() ,
                  &fEventBuf, bsize, split);
#else
     Char_t name[50];
     sprintf(name,"%s-EventBuf",GetName());
     tree->Branch(name, fEventBuf->ClassName() ,
                  &fEventBuf, bsize, split);
#endif
   }
}

//___________________________________________________________________________
void CDCRawData::SetBranch(TTree *tree)
{   
   fTree=tree;
   if( fEventBuf ) {
     Char_t name[50];
//
// 2001/09/14  Reverted to old branch name to use legacy unpacked data.
#if 0
     sprintf(name,"%s",fEventBuf->GetName());
#else
     sprintf(name,"%s-EventBuf",GetName());
#endif
     fBranch=tree->GetBranch(name);
#if 1
     fBranch->SetAddress(&fEventBuf);
#else
     if (fBranch) fBranch->SetAddress(&fEventBuf);
#endif
   }
}

